/*

File:          Newget.prg
Authors:       Sirius Software Development, Inc.
               Steve Straley
Compile:       RMake NewGet

Notes:  This file is part of the book:

    "Straley's Object Oriented Programming with Clipper/Class(y)"

    Published by Sirius Publishing

                            Sirius Publishing
             a division of Sirius Software Development, Inc.
                     564 Mission Street - Suite 343
                         San Francisco, CA 94105
                              415-399-9810
                              800-4-SIRIUS
                           415-399-9857 (fax)

    This example is used to illustrate the advantages of objects in their
    ability to co-exist with one another to achieve a total effect.  Too
    many books deal with tricks and not with the motivational factors
    needed to convince people to learn OOPs and to get ready for Visual
    Objects.  This books deals with those issues and more by focussing on
    Class(y) and Clipper to make a total package.

    In addition, all variables have been defined as they might appear
    in Visual Objects for Clipper based on the syntax seen at various
    demos and public locations.  The header file, PTVERBS.CH, is the
    most recent version of the header files from "Straley's Clipper 5.2
    Power Tools" published by Random House.

    Both source code for the Power Tools Library and the Power Tools Book
    is available from:

                    Sirius Software Development, Inc.
                     564 Mission Street - Suite 343
                         San Francisco, CA 94105
                              415-399-9810
                              800-4-SIRIUS
                           415-399-9857 (fax)


*/

#define CLIPPER   // User Clipper Standard stuff and not Power Tools!


#include "PTVerbs.ch"  // Power Tools Header file!
#include "PTValue.ch"  // Power Tools Miscelaneous Header File
#include "PTInkey.ch"  // Power Tools ASCII file

#include "Getexit.ch"  // CA-Clipper Exit File

#ifdef pTEST

  function Main()

    local cFilename            as char
    local getlist := {}        as array
    local cType   := space(10) as char
    local aDo     := {}        as array   // Array of behaviors for GET

    use Filenos new index Filenos   // This is based on file name order

    cFileName := Filenos->( dbgoto(0), fieldget( fieldpos( "file_name" ) ) )
    Filenos->( dbgoto(1) )

    // Add stuff to the behaviors

    // First, add the SEEKing mechanism
    aadd( aDo, {|cSeek| Filenos->( dbseek( cSeek ) )} )

    // Now, add a fully configured TBROWSE object with appropriate
    // Columns.  This is a complex expression so, to see it's impact,
    // disect the inner-most expression first, then resolve outward
    aadd( aDo, tbrowsedb(10,40,20,55):addcolumn( ;
               tbcolumnnew( ,{|| Filenos->file_name}))  )

    // This is the mechanism should a selection be made.
    aadd( aDo, {|| cFileName := Filenos->file_name} )

    cls
    @ 10,10 say "Enter Name:" get cFileName ;
            send reader := {|oGet, nSecs| PullDownReader(oGet, aDo, nSecs)}
    @ 11,10 say "      Type:" get cType
    read

    ? cFileName

    VOID

#endif

/*

Function:           PullDownReader( <oGet>, <aBehavior> [, <nSecs>] )
Parameters:         <oGet> :        Clipper GET object
                    <aBehavior> :   Array of behaviors and things
                    <nSecs> :       Number of seconds before timeout
Returns:            NIL
Purpose:            This function is a combination pull-down reader
                    that mixes a TBROWSE class and a GET class to
                    perform a unique behavior.  The array <aBehavior>
                    should consist of three items:

  aBehavior[1] -> code block containing the "lookup operation" within
                  the TBROWSE loop.  Typically, as the GET buffer contain
                  items, they will be passed to this code block.  Here,
                  if browsing off of a database, the item will be "seeked"
                  in the work area.  A different algorytm needs to be
                  worked out should this be an array.
  aBehavior[2] -> TBROWSE object with all components set including
                  TBCOLUMN object(s)
  aBehavior[3] -> code block for "assigning a value" should the ENTER
                  key be pressed within the BROWSING area on the screen

  The behavior of the PullDownReader() operation is simple.  The TAB and
  DOWN arrow keys from within the GET now go into the BROWSING thing.
  There, the cursor keys work within the area like a browse.  ESC always
  goes one level back so does the SHIFT_TAB keys.  Pressing the ENTER
  key from within the GET will not make a selection: you have to be in
  the browsing area to do this.

*/

*******************

function PullDownReader( oGet, ; // Standard modal read of a single GET
                    aBehavior, ; // Array of TBROWSE stuff and things
                      nSeconds ) //% 007 This is the number of seconds

// Manifest constants for the ComboBox/PullDown box defined herein
#define pLOCATOR aBehavior[1]
#define pWIDTH (oBrow:nRight - oBrow:nLeft - 2)
#define pBROWSER aBehavior[2]
#define pSELECTION aBehavior[3]

  local oBrow   as object
  local nKey    as int    //% 007 Holds the value of the key pressed
  local cScreen as char
  local nTop    as int
  local nLeft   as int
  local nBottom as int
  local nRight  as int

  DEFAULT nSeconds TO 0

  if ( getprevalidate( oGet ) )  // Read the GET if the WHEN condition
                                 // is satisfied

    // load up the browse
    oBrow   := pBROWSER

    // now set the screen coordinates for this
    nTop    := oBrow:nTop-1
    nLeft   := oBrow:nLeft-1
    nBottom := oBrow:nBottom+1
    nRight  := oBrow:nRight+1


    // Activate the GET for reading
    oGet:setFocus()

    // turn on the data area
    dispbegin()
    cScreen := savescreen( nTop, nLeft, nBottom, nRight )
    dispbox(nTop, nLeft, nBottom, nRight, pDBAR )

    // display the contents if there are any
    if !empty( oGet:buffer )
      devpos( oBrow:nTop-1, oBrow:nLeft+1 )
      devout( padr( alltrim(oGet:buffer), pWIDTH, chr(205) ) )
    endif

    // start us at the first element in the list
    if eval( pLOCATOR, alltrim(oGet:buffer) )
      oBrow:refreshall()
      while !( oBrow:stabilize() )
      enddo
    else
      oBrow:gotop()
    endif
    oBrow:hilite()
    oGet:killfocus()


    oBrow:configure()
    while !( oBrow:stabilize() )
    enddo
    oGet:setFocus()
    dispend()

    while ( oGet:exitState == GE_NOEXIT )

      if ( oGet:typeOut )            // Check for initial typeout
         oGet:exitState := GE_ENTER
      endif

      // Apply keystrokes until exit
      while ( oGet:exitState == GE_NOEXIT )
        if !empty( nKey := inkey(nSeconds) )  //% 007 Adding the timeout
          PullDownKey( oGet, nKey, oBrow, aBehavior )
        else
          exit
        endif

      enddo

      // Disallow exit if the VALID condition is not satisfied
      if ( !GetPostValidate( oGet ) )
        oGet:exitState := GE_NOEXIT
      endif
    enddo

    oGet:killFocus()   // De-activate the GET
    restscreen( nTop, nLeft, nBottom, nRight, cScreen )

  endif

  return( NIL )

/*

Function:           PullDownKey( <oGet>, <nKey>, <oBrow>, <aBehavior> )
Parameters:         <oGet> :        Clipper GET object in focus
                    <nKey> :        ASCII value of key pressed
                    <oBrow> :       Clipper TBROWSE object
                    <aBehavior>     Array of behaviors

Returns             NIL

Purpose:            This is the function that really uses both the GET
                    GET object and the TBROWSE object to combine one
                    effect.

*/

********************

static function PullDownKey( oGet, ;     // Get object having focus
                             nKey, ;     // ASCII value of key pressed
                            oBrow, ;     // The browsing object
                            aBehavior )  // Array of the behaviors

  static cSearch := ""      as char     // searching character

  local cKey := chr( nKey )         as char  // CHR of the key pressed
  local bKeyBlock                   as block // holds temp assigned key block
  local nNewKey                     as int   // holds the oBrow inkey value
  local nRow := oGet:row            as int
  local nCol := oGet:col + oGet:pos as int

  //% 002  Check the file-wide static to see if allows for the set
  //       key stack to be checked

  if !( ( bKeyBlock := setkey( nKey ) ) == NIL )
    GetDoSetKey( bKeyBlock, oGet )
    return( NIL )                                      // NOTE
  endif

  //% 005 This function returns a logical true (.T.) if the key pressed
  //      was part of the massive DO CASE test of "normal" keystrokes

  if GetDefKeys( oGet, nKey )
    // In case the movement of the cursor dictates that it is
    // in a new position
    nRow := oGet:row
    nCol := oGet:col + oGet:pos - 1
  else

    // if the keystroke was not a "default keystroke", then a test will
    // be made to see if the key was something "else".  The previous
    // function, "GETDEFKEYS" should be modified to include the TAB
    // and DOWN ARROW keys and placed inside of the modified GETSYS.PRG
    // (as it is done in
    //
    //            HELIOS - Mouse Aware Systems
    //     Published by Sirius Software Development, Inc.
    //                  1-800-4-SIRIUS
    //
    // This way, the "default keys" can be stored in one place and not
    // duplicated as it appears in the GETSYS.PRG file that comes with
    // Clipper.

    do case
    case (nKey == pTAB .or. nKey == pDOWN_ARROW)

      while !( oBrow:stabilize() )
      enddo

      PROCESS   // User Defined Command in POWER TOOLS 5.2

        do case
        case (nNewKey := inkey(0)) == pDOWN_ARROW
          oBrow:down()
        case nNewKey == pPGDN
          oBrow:pageDown()
        case nNewKey == pPGUP
          oBrow:pageUp()
        case nNewKey == pUP_ARROW
          oBrow:up()
        case nNewKey == pSHIFT_TAB   // Go back to the get
          nCol--
          exit
        case nNewKey == pTAB         // Move onto the next GET w/o change
          oGet:exitState := GE_ENTER
          exit

        case nNewKey == pENTER       // Make a selected and move
          oGet:killfocus()
          eval( pSELECTION )
          oGet:display()
          oGet:setfocus()
          oGet:exitState := GE_ENTER
          exit

        case nNewKey == pESC         // ESC out and go back
          nCol--
          exit

        endcase

        while !( oBrow:stabilize() )
        enddo

      END PROCESS    // User Defined Commands in POWER TOOLS

    otherwise
      if ( nKey >= 32 .and. nKey <= 255 )

        if ( oGet:type == "N" .and. ( cKey == "." .OR. cKey == "," ) )
          oGet:toDecPos()
        else

          if ( set( pINSERT ) )
            oGet:insert( cKey )
          else
            oGet:overstrike( cKey )
          endif

          if ( oGet:typeOut )
            if ( set( pBELL ) )
              qqout( chr(7) )
            endif
            if ( !set( pCONFIRM ) )
              oGet:exitState := GE_ENTER
            endif
          endif

        endif

      endif
    endcase
  endif

  cSearch := alltrim( oGet:buffer )
  devpos( oBrow:nTop-1, oBrow:nLeft+1 )
  devout( padr( cSearch, pWIDTH, chr(205) ) )

  if eval( pLOCATOR, cSearch )      // seek the contents of the string
    oBrow:refreshall()
    while !( oBrow:stabilize() )
    enddo
  endif

  oBrow:hilite()
  setpos( nRow, nCol )

  VOID

/* -----------------------

Function:           GetDefKeys( <oGet>, <nKey> )
Parameters:         <oGet> - Get object that is currently in focus
                    <nKey> - Ascii value of the key pressed
Returns:            <lSuccess>
Note:               This function will return a logical true (.T.)
                    value should the key pressed be a "standard"
                    movement/cursor key.  It will also perform the
                    appropriate message.

*/

********************

function GetDefKeys( oGet, ;   // Object of the GET that has input focus
                     nKey )    // ASCII value of the key pressed

  local lMember := pTRUE as logical // Default is a true to be altered
                                    // in the OTHERWISE at the end

  do case
  case ( nKey == pUP_ARROW )
    oGet:exitState := GE_UP

  case ( nKey == pSHIFT_TAB )
    oGet:exitState := GE_UP

  case ( nKey == pENTER )
    oGet:exitState := GE_ENTER

  case ( nKey == pESC )
    if ( set( pESCAPE ) )
      oGet:undo()
      oGet:exitState := GE_ESCAPE
    endif

  case ( nKey == pPGUP )
    oGet:exitState := GE_WRITE    //      wide static variable

  case ( nKey == pPGDN )
    oGet:exitState := GE_WRITE    //      wide static variable

  case ( nKey == pCTRL_HOME )
    oGet:exitState := GE_TOP

  case ( nKey == pCTRL_END )
    oGet:exitState := GE_BOTTOM

  case ( nKey == pCTRL_W )
    oGet:exitState := GE_WRITE

  case ( nKey == pINS )
    setcursor( if(set( pINSERT, ;
                       !set( pINSERT ) ), 1, 3 ) )  //% 001

  case ( nKey == pCTRL_U )
    oGet:undo()

  case ( nKey == pHOME )
    oGet:home()

  case ( nKey == pEND )
    oGet:end()

  case ( nKey == pRIGHT_ARROW )
    oGet:right()

  case ( nKey == pLEFT_ARROW )
    oGet:left()

  case ( nKey == pCTRL_RIGHT_ARROW )
    oGet:wordRight()

  case ( nKey == pCTRL_LEFT_ARROW )
    oGet:wordLeft()

  case ( nKey == pBKSP )
    oGet:backSpace()

  case ( nKey == pDEL )
    oGet:delete()

  case ( nKey == pCTRL_T )
    oGet:delWordRight()

  case ( nKey == pCTRL_Y )
    oGet:delEnd()

  case ( nKey == pCTRL_BKSP )
    oGet:delWordLeft()

  otherwise
    lMember := pFALSE

  endcase

  return( lMember )

// End of File: PullDown.prg
